package gkdtree;

/**
 *
 * Main class of the filtering process.
 * This is a high level overview of the process.
 * 
 * On initialization the file is load and the file-octree is generated.
 * (Which is a layer between the data on the disk and the algorithm.)
 * 
 * help: importance sampling 
 *   - http://www.hss.caltech.edu/~mshum/gradio/simulation.pdf
 *   - http://mpdc.mae.cornell.edu/Courses/UQ/ImportanceSampling.pdf
 *   - http://www.youtube.com/watch?v=7TybpwBlcMk&feature=relmfu
 *   - http://www.youtube.com/watch?v=S3LAOZxGcnk&feature=relmfu
 * 
 * normally 1/n
 * 
 * @author DarkMatt
 *
 */

public class FilterMain {

	public int dimension = 3;
	
	private FileOctree data;
	private GKDTree gkdt;
	//private Boundary selection;
	private int[] selected_cells;
	private float[] points;
	
	private int[] data_size = new int[]{1200, 1200};
	
	private static final float SIGMA_SCALE = 250f;
	private float used_sigma = SIGMA_SCALE;
	
	public FilterMain(String filename, int[] size) {
		
		data_size = size;
		used_sigma = size[0];
		
		/*
		 * 
		 * On sigma:
		 * 
		 * 2 standard deviations! range/domain -> how to set?
		 * 
		 * magical paper values:
		 * sigma_s = 1/sqrt(11)
		 * sigma_b = 3/sqrt(11)
		 * 
		 */
		
		data = new FileOctree();
		//data.generateNewShort("data/sphere_100x100x100", new int[]{100, 100, 100}, true);
		//data.initFile(-1, "data/sphere_100x100x100");
		
		/*
		data.generateNewShort("data/testdata_100x100x100", new int[]{100, 100, 100}, true);
		data.initFile(-1, "data/testdata_100x100x100");
		selected_cells = data.queryArea(new Boundary(new int[]{0, 0, 0}, new int[]{100, 100, 100}));
		*/
		
		//NO SPHERE! (this only works 3d.. atm)
		//data.generateNew2DChecker("data/testdata_100x100", new int[]{100, 100});
		//checker generated
		/*data.generateNew2DChecker("data/testdata_100x100", data_size);
		data.initFile(-1, "data/testdata_100x100");
		gkdt = new GKDTree("data/testdata_100x100", dimension, data);*/
		
		
		//file
		data.generateNew2DFile("data/testimg_1000x1000", filename, data_size);
		data.initFile(-1, "data/testimg_1000x1000");
		gkdt = new GKDTree("data/testimg_1000x1000", dimension, data, used_sigma);
		//file end
		
		selected_cells = data.queryArea(new Boundary(new int[]{0, 0}, data_size));
		
		//data.printTree();
		//gkdt.initTree(30.0f); 
		//TODO: debug small sigma with test-data - probably if node is empty!
		//should this ever happen? - CHECK!!!
		
		//ArrayList<GKDNode.Result> tmp = gkdt.queryIntoTree(new float[]{500.0f, 500.0f, 500.0f}, 10000000.0f, 100);
		
		/*System.out.println("Printing query results (" + tmp.size() + "):");
		for (int i=0; i<tmp.size(); i++) {
			float[] tmp_p = tmp.get(i).node.pos;
			System.out.println(tmp_p[0] + "/" + tmp_p[1] + "/" + tmp_p[2] + " weight: " + tmp.get(i).weight);
		}*/
		
		//points = data.getPointsOctree(32000);
		
		//data.debug_draw();
		
		//System.out.println(points.length);
		//for (int i=0; i<points.length; i++) System.out.println(points[i]);
		
		// TEST-case failing...!
		/*int[] result = data.queryArea(new Boundary(new int[]{0, 51, 0}, new int[]{100, 49, 100}));
		if (result.length > 0) {
			System.out.println("Area query result: ");
			for (int i=0; i<result.length; i++) {
				System.out.print(result[i] + ", ");
			}
			System.out.println(result[result.length-1]);
		}*/
		
	}
	
	public void initTree(float sigma_s) {
		float sig_s = 1.0f/(float)Math.sqrt(11);
		gkdt.initTree((sigma_s * sig_s) / used_sigma); //used_sigma is only size... e.g. 100
	}
	
	/**
	 * Applies the bilateral filter with use of Gaussian kd trees.
	 * 1. Initialize kd-tree
	 * 2. Downsample the data-set: save the result to the kd-tree leaves
	 * 3. Blur the kd-leaves: Query the leaves and store the result in the center.
	 * 4. Upsample the kd-leaves: splat the values of the nodes out to the original data points.
	 * 
	 * sigma_s spatial
	 * sigma_r range
	 * 
	 * @param sigma_s
	 * @param sigma_r
	 */
	public void apply_Bilateral_Filter(float sigma_s, float sigma_r) {
		
		float sig_s = 1.0f/(float)Math.sqrt(11);
		float sig_b = 3.0f/(float)Math.sqrt(11);
		
		//float[] data1 = data.get;
		
		// first scale to have sigma == 1
		gkdt.scale(sigma_s, sigma_r);
		//downsample with sig_s
		gkdt.downsample(sig_s); //used to be sig_s
		
		System.out.println("____________________________ BLUR PASS ____________________________");
		
		//blur with sig_b
		gkdt.blur(sig_b);
		
		System.out.println("____________________________ UPSAMPLE PASS ____________________________");
		
		//upsample with sig_s
		gkdt.upsample(sig_s);
		
		gkdt.undoScale();
		
		//float[] data2 = gkdt.get_drawingStructure();
		
		System.out.println("Done.");
		
		System.out.println();
		
		//result should be generated in upsample...
		
		//get-points does not work on leaves?
		
	}
	
	public float[] getKDDrawing() {
		return gkdt.get_drawingStructure();
	}
	
	public float[] getKDDrawingUV() {
		return gkdt.get_drawingStructureUV();
	}
	
	public float[] getKDDrawingColor() {
		return gkdt.get_drawingStructureColor();
	}
	
	public float[] getPoints() {
		points = data.getPointsOctree2D(0, selected_cells);
		
		System.out.println("num points: " + points.length/3);
		
		return points;
	}
	
	public void setSelection(Boundary b) {
		//selection = b;
		selected_cells = data.queryArea(b);
		
		System.out.print("Cell indices: ");
		for (int i=0; i<selected_cells.length-1; i++)
			System.out.print(selected_cells[i] + ", ");
		System.out.println(selected_cells[selected_cells.length-1] + ".");
	}
	
	public float[] getTreeLeaves() {
		return gkdt.get_Leaves();
	}
	
	public float[] getTreeLeavesColor() {
		return gkdt.get_LeavesColor();
	}
	
}
