package gkdtree;

import java.util.ArrayList;
import java.util.Arrays;

/**
 *
 * Leaf nodes of the kd-tree.
 * 
 * @author DarkMatt
 *
 */

public class GKDLeafNode extends GKDNode {

	public float[] pos;
	public	float value;
	public float weight;
	
	public GKDLeafNode(float[] position) {
		this.pos = position;
		value = 0.0f;
		weight = 1.0f;
	}
	
	public void applyScale(float[] scale) {
		this.scale = scale.clone();
		for (int i=0; i<pos.length; i++) pos[i] *= scale[i];
	}
	
	public void undoScale() {
		for (int i=0; i<pos.length; i++) pos[i] /= scale[i];
		Arrays.fill(scale, 1.0f);
	}
	
	public void applyValue() {
		pos[pos.length-1] = value/weight;//GKDTree.NUM_SAMPLES;//weight;
	}
	
	public void query(float[] pos, float sigma, int num_samples, 
			ArrayList<Result> results, float p, ArrayList<GKDNode> nodes) {
		
		//if (p == 0) p += 0.0001;
		
		float distance = Distance.euclidean(pos, this.pos);
		float correctP = (float)Math.exp(-(distance*distance)/(2.0f*sigma)); // in paper distance*distance
		
		/*System.out.println(
				((float)num_samples*correctP/p)
				+ " {d: " + distance + ", " + p + ", " + correctP + "} " +
						" (" + this.pos[0] +"/"+ this.pos[1] +"/"+ this.pos[2] +"): " + num_samples);*/
		
		results.add(new Result(this, (float)num_samples*correctP/p));
		
	}
	
}
