package gkdtree;

import gkdtree.GKDNode.Result;

import java.util.ArrayList;


/**
 *
 * Implementation of the Gaussian kd-tree. Supports the initialization and queries
 * as well as dispaly functionality.
 *
 * "complicated theories only seem elegant in theory"
 * 
 * @author DarkMatt
 *
 */

//TODO implement scaling! scale method scale[]

//TODO range/domain weighting/scaling needs to depend on the data-set size

public class GKDTree {
	
	private int debug_node_count = 0;
	public static final int NUM_SAMPLES = 150; //150
	private static final boolean DEBUG = true;
	
	private float[] scale;
	private float[] vol_size;
	
	/**
	 * 
	 * @author DarkMatt
	 *
	 * A struct for the return type of getBBMultiList().
	 *
	 */
	private class BBReturnStruct {
		public ArrayList<Float> points;
		public GKDNode node;
		public int prev_size;
	}
	
	public static final int MAX_LEVEL = 10000;
	
	private int dimension;
	private FileOctree octree;
	
	private ArrayList<GKDNode> nodes;
	
	float[] drawing_structure;
	
	private float sigma_sq = 1.0f;
	private float sigma_spatial = 1.0f;
	
	public GKDTree(String filename, int dimension, FileOctree data, float scale_R) {
		
		this.dimension = dimension;
		this.octree = data;
		
		nodes = new  ArrayList<GKDNode>();
		
		scale = new float[dimension];
		for (int i=0; i<dimension; i++) scale[i] = 1.0f;
		
		vol_size = new float[dimension];
		for (int i=0; i<dimension-1; i++) vol_size[i] = (float)octree.getBoundary().getSize()[i];
		vol_size[dimension-1] = scale_R/octree.getLimit();
		
	}
	
	/**
	 * Starts the tree initialization method.
	 * 
	 * @param sigma
	 */
	public void initTree(float sigma) {
		
		sigma_sq = sigma * sigma;
		
		// the boundary is where the array is indexed
		Boundary b = octree.getBoundary();
		
		if (DEBUG) {
			//System.out.println("Initial size: ");
			//b.printBoundary("Total volume");
		}
		
		/*GKDNode test = new GKDInnerNode(0,0,0,0);
		GKDNode test2 = new GKDLeafNode(null);
		
		System.out.println(GKDInnerNode.class.isInstance(test));
		System.out.println(test2.getClass());*/
		
		split(true, -1, 0, 2, b, 0, octree.getLimit(), null);
		
		System.out.println("Done generating kd-tree.");
		
	}
	
	/**
	 * Recursion for tree generation
	 * 
	 * leaves: store one point each
	 * 
	 * Boundary.dimension = dimension -1
	 * 
	 * @param left true ... left leaf, false ... right leaf
	 * @param parent parent node index
	 * @param level 
	 * @param previous_bb_elements number of partitions of the file-octree
	 * @param b current boundary of the recursion
	 * @param lower current value range of the recursion
	 * @param upper current value range of the recursion
	 * @param elements a list of elements (this is only not null if the size is small enough)
	 */
	private void split(boolean left, int parent, int level, int previous_bb_elements, Boundary b, float lower, float upper, ArrayList<Float> elements) {
		
		/*System.out.println("=================================" 
				+ " Node: " + debug_node_count + "/" + level + ", l/u: " + lower + "/" + upper + 
				" =================================");
		b.printBoundary("node "+debug_node_count);
		System.out.println(lower+"/"+upper);*/
		debug_node_count++;
		
		if (level > MAX_LEVEL) {
			System.out.println("maxlevelbreak");
			return;
		}
		
		int this_node = nodes.size(); //not added yet
		
		/*if (b.getUpper()[0] > 100 || b.getUpper()[1] > 100) {// || b.getUpper()[2] > 100) {
			System.out.println("FRRRRRRRRRRRRRRRRREAKING EEEEEEEEEEEEEEERRRRRRRRRRRRROR");
			//THIS ALLOWS A WONDERFUL INFINITE LOOP :)
			return;
		}*/
		
		//gets the bounding-box (depends on the elements in this node)
		GKDNode node;
		int list_partition_count;
		ArrayList<Float> element_list = elements;
		if (previous_bb_elements > 1) {
			
			// PROBLEM: large data chunks don't get smaller due to value-range splits!
			
			BBReturnStruct bb = getBBMultiList(previous_bb_elements, b, lower, upper);
			node = bb.node;
			list_partition_count = bb.prev_size;
			if (list_partition_count == 1) {
				element_list = bb.points;
			}
			
			//System.out.println("Multi");
		} else {
			
			//ERROR: generated here
			
			node = getBBList(elements, lower, upper);
			list_partition_count = 1;
			
			//System.out.println("Single");
		}
		
		//System.out.println(lower+"/"+upper); //weird...
		
		//add childs
		if (parent != -1) {
			if (left) nodes.get(parent).setLeftChild(this_node);
			else nodes.get(parent).setRightChild(this_node);
		}
		
		// LEAF NODES
		if ( !GKDInnerNode.class.isInstance(node) ) {
			nodes.add(node);
			//System.out.println("Leaf NODE (class): " + node.getClass());
			return; //sigma break
		}
		
		int tmp_d = node.dimension;
		float tmp_cut = node.cut;
		
		GKDInnerNode tmp = (GKDInnerNode)node; //has to be GKDInnerNode (if statement above)
		if (tmp_d < dimension-1) {
			tmp.min = b.getLower()[tmp_d];
			//can not be -> probability towards zero for non inf
			//if (b.getLower()[tmp_d] == 0) tmp.min = Float.NEGATIVE_INFINITY;
			tmp.max = b.getUpper()[tmp_d];
			//if (b.getSize()[tmp_d] >= octree.getBoundary().getSize()[tmp_d]) tmp.max = Float.POSITIVE_INFINITY;
		} else {
			tmp.min = lower;
			//if (lower == 0) tmp.min = Float.NEGATIVE_INFINITY;
			tmp.max = upper;
			//if (upper >= octree.getLimit()) tmp.max = Float.POSITIVE_INFINITY;
		}
		nodes.add(tmp);
		
		//sort the index array
		//System.out.println("Cut AT ("+tmp_d+"): " + tmp_cut);
		
		ArrayList<Float> left_elements = null;
		ArrayList<Float> right_elements = null;
		if (element_list != null) {
			left_elements = new ArrayList<Float>();
			right_elements = new ArrayList<Float>();
			for (int i=0; i<element_list.size()/dimension; i++) {
				float th = element_list.get(dimension*i + tmp_d);
				for (int d=0; d<dimension; d++) {
					if (th <= tmp_cut) 
						left_elements.add(element_list.get(dimension*i + d));
					else 
						right_elements.add(element_list.get(dimension*i + d));
				}
			}
		}
		
		//TODO: boundaries (scale cut to array...)
		
		//p problem if split along 4th dimension...
		Boundary[] b_new = new Boundary[2];
		float left_lower = lower;
		float left_upper = upper;
		float right_lower = lower;
		float right_upper = upper;
		if (tmp_d != dimension -1) {
			b_new = b.cut(tmp_d, tmp_cut);
			//b_new[0].printBoundary("left");
			//b_new[1].printBoundary("right");
		} else {
			b_new[0] = b;
			b_new[1] = b;
			left_upper = tmp_cut;
			right_lower = tmp_cut;
			
			if (right_lower > right_upper) {
				System.out.println("Not supposed to happen: " + right_lower + "/" + right_upper);
			}
			
		}
		
		split(true, this_node, level + 1, list_partition_count, b_new[0], left_lower, left_upper, left_elements); //lower-bound
		split(false, this_node, level + 1, list_partition_count, b_new[1], right_lower, right_upper, right_elements); //bound-upper
		
	}
	
	/**
	 * Returns the split node for a list of points.
	 * 
	 * Return null if sigma condition failure
	 * 
	 * This method does not have any knowledge of the regarded area for the BB-generation!
	 * This is necessary for min/max!
	 * 
	 * @param points
	 * @return GKDNode
	 */
	private GKDNode getBBList(ArrayList<Float> points, float lower, float upper) {
		GKDInnerNode split = null;
		
		//was this the mistake? (created lower > upper cases)
		//float[][] bb = getBB(points, null, 0, octree.getLimit());
		float[][] bb = getBB(points, null, lower, upper);
		float smallest[] = bb[0];
		float largest[] = bb[1];
		
		//calculate diagonal
		float distance = 0.0f;
		for (int d=0; d<dimension; d++) {
			float val = (largest[d]-smallest[d]);
			
			if (d == dimension - 1) val *= vol_size[d];
			
			distance += val*val;
		}
		//compare distance of diagonal to sigma
		if (distance < sigma_sq || points.size() <= 1) {
			//System.out.println("sigma break (d/sig): " + distance + "/" + sigma_sq);
			float[] pos = new float[dimension];
			for (int d=0; d<dimension; d++) {
				pos[d] = smallest[d] + (largest[d]-smallest[d])/2.0f;
			}
			return new GKDLeafNode(pos);
		}
		
		//otherwise: get largest bounding box dimension and split
		float value = 0;
		int largest_dim = -1;
		for (int d=0; d<dimension; d++) {
			float val = Math.abs(largest[d]-smallest[d]);
			
			//scales the value range...
			if (d == dimension - 1) val *= vol_size[d];
			
			if (val > value) {
				value = val;
				largest_dim = d;
			}
		}
		
		if (largest_dim == dimension - 1) value /= vol_size[dimension - 1];
		
		//split_at+1 : this is required for a checker board to work correctly...
		float split_at = smallest[largest_dim] + (largest[largest_dim]-smallest[largest_dim])/2.0f;
		split = new GKDInnerNode(largest_dim, 0, 0, split_at);
		return split;
	}
	
	/**
	 * Returns the split node for a large octree sub-region.
	 * 
	 * Returns:
	 *  - split-node
	 *  - number of splits for processing the list of points
	 *  - list of points if number of splits is 1
	 *  
	 *  (The most crucial splitting logic is done in FileOctree - PointList)
	 *  
	 *  Boundary b and lower/upper are the boundary of the considered area.
	 *  
	 * @param previous_points_size
	 * @param b
	 * @param lower
	 * @param upper
	 * @return BBReturnStruct
	 */
	private BBReturnStruct getBBMultiList(int previous_points_size, Boundary b, float lower, float upper) {
		BBReturnStruct out = new BBReturnStruct();
		GKDInnerNode split = null;
		
		int curr_size = octree.initPointListLoad(b);
		
		ArrayList<Float> points = null;
		float[][] bb = null;
		
		//System.out.println(curr_size);
		
		for (int i=0; i<curr_size; i++) {
			points = octree.getNextPointList();
			bb = getBB(points, bb, lower, upper);
		}
		//System.out.println( bb[0]);
		float smallest[] = bb[0];
		float largest[] = bb[1];
		
		if (DEBUG) {
			//b.printBoundary("l'horreur");
			//System.out.println("DIMENSION: " + dimension);
			//System.out.println("Freaking trashvalue: " + largest[dimension-1]);
		}
		
		//calculate diagonal
		float distance = 0.0f;
		for (int d=0; d<dimension; d++) {
			
			float val = (largest[d]-smallest[d]);
			
			if (d == dimension - 1) val *= vol_size[d];
			
			distance += val*val;
		}
		
		//System.out.println("dist: " + distance);
		
		//compare distance of diagonal to sigma
		if (distance < sigma_sq ||  Float.isInfinite(distance)) {
			//System.out.println("sigma break (d/sig): " + distance + "/" + sigma_sq);
			float[] pos = new float[dimension];
			for (int d=0; d<dimension; d++) {
				pos[d] = smallest[d] + (largest[d]-smallest[d])/2.0f;
			}
			out.node = new GKDLeafNode(pos);
			return out;
		}
		
		//TODO PENALTY FOR SHORT
		
		//otherwise: get largest bounding box dimension and split
		float value = 0;
		int largest_dim = -1;
		for (int d=0; d<dimension; d++) {
			float val = Math.abs(largest[d]-smallest[d]);
			
			//scales the value range...
			if (d == dimension - 1) val *= vol_size[d];
			
			if (val > value) {
				value = val;
				largest_dim = d;
			}
		}
		
		if (largest_dim == dimension - 1) value /= vol_size[dimension - 1];
		
		//split_at+1 : this is required for a checker board to work correctly... (no..)
		float split_at = smallest[largest_dim] + (largest[largest_dim]-smallest[largest_dim])/2.0f;
		split = new GKDInnerNode(largest_dim, 0, 0, split_at);
		
		out.node = split;
		out.prev_size = curr_size;
		if (curr_size > 1) out.points = null;
		else out.points = points;
		
		return out;
	}
	
	/**
	 * Returns the bounding box for a list of points.
	 * 
	 * @param points
	 * @param initial
	 * @param lower
	 * @param upper
	 * @return float[][]
	 */
	private float[][] getBB(ArrayList<Float> points, float[][] initial, float lower, float upper) {
		
		float smallest[] = new float[dimension];
		float largest[] = new float[dimension];
		if (initial == null) {
			for (int i=0; i<dimension; i++) {
				smallest[i] = Float.MAX_VALUE;
				largest[i] = -Float.MAX_VALUE;
			}
			//smallest[dimension-1] = lower;
			//largest[dimension-1] = upper;
		} else {
			for (int i=0; i<dimension; i++) {
				smallest[i] = initial[0][i]; //TODO: is this correct?
				largest[i] = initial[1][i];
			}
		}
		
		//get bounding box in all directions
		//PROBLEM: sometimes not overwritten...
		for (int i=0; i<points.size()/dimension; i++) {
			if (	points.get(dimension*i + dimension - 1) >= lower &&
					points.get(dimension*i + dimension - 1) < upper) {
				for (int d=0; d<dimension; d++) {
					float val = points.get(dimension*i + d);
					if (val < smallest[d]) smallest[d] = val;
					if (val > largest[d]) largest[d] = val;
				}
			}
		}
		
		float[][] bb = new float[2][dimension];
		bb[0] = smallest;
		bb[1] = largest;
		
		if (DEBUG) {
			/*System.out.println("Bounding box:");
			for (int i=0; i<dimension-1; i++) {
				System.out.print(bb[0][i]+"/"+bb[1][i] + ", ");
			}
			System.out.println(bb[0][dimension-1]+"/"+bb[1][dimension-1]);*/
		}
		
		return bb;
	}
	
	/**
	 * Sets the scale parameters for the data-set.
	 * 
	 * @param sigma_s
	 * @param sigma_r
	 */
	public void scale(float sigma_s, float sigma_r) {
		
		//System.out.println("Doing scale ("+nodes.size()+")...");
		//sigma_spatial = sigma_s;
		//vol_size[dimension-1] *= sigma_spatial;
		
		int[] size = octree.getBoundary().getSize();
				
		
		// divided by sigma: small sigma => data-set larger
		for (int i=0; i<dimension-1; i++) {
			//first normalized then scald by inverse standard deviation
			scale[i] = (1.0f/(float)size[i]) / sigma_s;// / sigma_s;
		}
		scale[dimension-1] = (1.0f/octree.getLimit()) / sigma_r;// / sigma_r;
		
		System.out.println("Scale: ");
		System.out.println(scale[0] + "/" + scale[1] + "/" + scale[2]);
		
		for (int i=0; i<nodes.size(); i++) {
			nodes.get(i).applyScale(scale);
		}
		
	}
	
	public void undoScale() {
		//undo scale for drawing
		for (int i=0; i<nodes.size(); i++) {
			nodes.get(i).undoScale();
		}
	}
	
	/*
	 * I need to scale the values for the query process!
	 * Scale all position values
	 */
	
	/**
	 * For all data-points do a Gaussian query - save results to leaf nodes
	 * 
	 * Query results in a list of points. Color values are assigned to
	 * the points according to the query origin value.
	 * 
	 * @param sigma
	 */
	public void downsample(float sigma) {
		
		//System.out.println("started downsample...");
		
		int data_chunks = octree.initPointListLoad(octree.getBoundary());
		
		int samples = NUM_SAMPLES;//octree.getBoundary().getNumberOfElements();
		
		System.out.println("Samplenumber: " + samples);
		
		for (int dc=0; dc<data_chunks; dc++) {
			
			ArrayList<Float> data = octree.getNextPointList();
			
			// for all grid points
			for (int i=0; i<data.size()/dimension; i++) {
				
				float[] pos = new float[dimension];
				//already normalized (scaled 0-1)
				for (int d=0; d<dimension; d++) pos[d] = data.get(dimension*i+d)*scale[d];
				
				//System.out.println("query: " + pos[0] +"/"+ pos[1] +"/"+ pos[2]);
				
				ArrayList<Result> n = this.queryIntoTree(pos, sigma, samples);
				
				
				//int c = 0;
				for (int j=0; j<n.size(); j++) {
					
					GKDLeafNode node = n.get(j).node;
					if (!Float.isNaN(n.get(j).weight) && !Float.isInfinite(n.get(j).weight) &&
							n.get(j).weight > 0.001) {
						node.value += pos[dimension-1] * n.get(j).weight; // 
						//c ++;
						node.weight += n.get(j).weight; //NECESSARY? - it is sufficient to divide by MAX_SAMPLES!
					} else {
					//	System.out.println("ERROR: downsample");
					}
					
				}
				//node.value
				
				//node.value /= weight;
				
			}
			
		}		
		
		//TODO: apply changes!
		for (int i=0; i<nodes.size(); i++) {
			nodes.get(i).applyValue();
		}
		
		System.out.println("Downsampling done.");
		
	}
	
	/**
	 * Blur is a Query at every leaf node and propagates
	 * the value of the leaf node to adjacent leaf nodes.
	 * 
	 * This requires a second value at every leaf node, or
	 * a copy of the tree.
	 * 
	 * @param sigma
	 */
	public void blur(float sigma) {
		ArrayList<GKDLeafNode> leaves = getLeavesList();
		
		for (int i=0; i<leaves.size(); i++) {
			
			float[] pos = leaves.get(i).pos;
			
			//System.out.println("query: " + pos[0] +"/"+ pos[1] +"/"+ pos[2]);
			
			ArrayList<Result> n = this.queryIntoTree(pos, sigma, NUM_SAMPLES);
			
			for (int j=0; j<n.size(); j++) {
				GKDLeafNode node = n.get(j).node;
				if (!Float.isNaN(n.get(j).weight) && !Float.isInfinite(n.get(j).weight)) {
					node.value += n.get(j).weight * pos[dimension-1]; // 
					node.weight += n.get(j).weight; //NECESSARY?
				} else
					System.out.println("ERROR: blur");
			}
			
		}	
		
		//TODO: apply changes!
		for (int i=0; i<nodes.size(); i++) {
			nodes.get(i).applyValue();
		}
	}
	
	/**
	 * For all data-points a Query is executed.
	 * 
	 * The values are sliced to the data-points.
	 * 
	 * @param sigma
	 */
	public void upsample(float sigma) {
		int data_chunks = octree.initPointListLoad(octree.getBoundary());
		
		System.out.println("number of data chunks " + data_chunks);
		
		// for all data-chunks
		for (int dc=0; dc<data_chunks; dc++) {
			
			// list of points
			ArrayList<Float> data = octree.getNextPointList();
			
			// for all points
			for (int i=0; i<data.size()/dimension; i++) {
				
				float[] pos = new float[dimension];
				//already normalized (scaled 0-1)
				for (int d=0; d<dimension; d++) pos[d] = data.get(dimension*i+d)*scale[d];
				
				//System.out.println("query: " + pos[0] +"/"+ pos[1] +"/"+ pos[2]);
				
				// list of nodes
				ArrayList<Result> n = this.queryIntoTree(pos, sigma, NUM_SAMPLES);
				
				float result_value = 0.0f;
				float result_weight = 0.0f;
				for (int j=0; j<n.size(); j++) {
					
					if (!Float.isNaN(n.get(j).weight) && !Float.isInfinite(n.get(j).weight)) {
						result_value += n.get(j).weight * n.get(j).node.pos[dimension-1];//pos[dimension-1]; //  * pos[dimension-1]
						result_weight += n.get(j).weight;
					} else {
						System.out.println("no ok: " + n.get(j).weight);
					}
					
				}
				result_value /= result_weight;
				//scale back
				result_value /= scale[dimension-1];
				
				//System.out.println(result_value);
				
				//if (Float.isNaN(result_value)) //sometimes the values are nan.. (already in weight)
					//System.out.println(result_value);
				/*if (!Float.isNaN(result_value) && result_value != 0.0) {
					System.out.println("ok ("+pos[0]+"/"+pos[1]+"/"+pos[2]+"): " + result_value);
				} else {
					System.out.println("!ok ("+pos[0]+"/"+pos[1]+"/"+pos[2]+"): " + result_value);
				}*/
				
				// TODO: probably needs to be done differently!
				if (!Float.isNaN(result_value)) {
					//System.out.println(data.get(dimension*i+dimension-1) +" vs "+ result_value);
					
					data.set(dimension*i+dimension-1, result_value); //used to be add! :(
				} else
					System.out.println("Error - no write!");

			}
			
			// write to file?
			octree.writePointList(data);
			
		}
	}
	
	/**
	 * Starts a Gaussian query at the leaf node.
	 * (For details see paper.)
	 * 
	 * @param pos
	 * @param sigma
	 * @param num_samples
	 * @return ArrayList<GKDNode.Result>
	 */
	public ArrayList<GKDNode.Result> queryIntoTree(float[] pos, float sigma, int num_samples) {
		
		ArrayList<GKDNode.Result> results = new ArrayList<GKDNode.Result>();
		
		nodes.get(0).query(pos, sigma, num_samples, results, 1, nodes);
		
		return results;
		
	}
	
	/**
	 * Returns quads at the cut values of the kd-tree.
	 * (still not working 100% correctly... its drawn or all nodes)
	 * 
	 * @return float[]
	 */
	public float[] get_drawingStructure() {
		
		System.out.println("Drawing nodes size: " + nodes.size());
		
		drawing_structure = new float[4*3*nodes.size()];
		
		//a number of quads
		
		for (int i=0; i<nodes.size(); i++) {
			int d = nodes.get(i).dimension;
			float cut = (nodes.get(i).cut + 0.5f) / vol_size[d]; // 100.0f; //TODO scale?
			if (nodes.get(i).cut == 0.0) cut = - 10000; //nice hack to fix it :P (moves the invalid ones far away...)
			if (d == dimension-1) cut /= (float) Short.MAX_VALUE / vol_size[d]; // 100.0f;//
			
			if (d == 0) {
				drawing_structure[4*3*i + 0] = cut; drawing_structure[4*3*i + 1] = 0.0f; drawing_structure[4*3*i + 2] = 0.0f;
				drawing_structure[4*3*i + 3] = cut; drawing_structure[4*3*i + 4] = 0.0f; drawing_structure[4*3*i + 5] = 1.0f;
				drawing_structure[4*3*i + 6] = cut; drawing_structure[4*3*i + 7] = 1.0f; drawing_structure[4*3*i + 8] = 1.0f;
				drawing_structure[4*3*i + 9] = cut; drawing_structure[4*3*i + 10] = 1.0f; drawing_structure[4*3*i + 11] = 0.0f;
			}
			if (d == 1) {
				drawing_structure[4*3*i + 0] = 0.0f; drawing_structure[4*3*i + 1] = cut; drawing_structure[4*3*i + 2] = 0.0f;
				drawing_structure[4*3*i + 3] = 0.0f; drawing_structure[4*3*i + 4] = cut; drawing_structure[4*3*i + 5] = 1.0f;
				drawing_structure[4*3*i + 6] = 1.0f; drawing_structure[4*3*i + 7] = cut; drawing_structure[4*3*i + 8] = 1.0f;
				drawing_structure[4*3*i + 9] = 1.0f; drawing_structure[4*3*i + 10] = cut; drawing_structure[4*3*i + 11] = 0.0f;
			}
			if (d == 2) {
				drawing_structure[4*3*i + 0] = 0.0f; drawing_structure[4*3*i + 1] = 0.0f; drawing_structure[4*3*i + 2] = cut;
				drawing_structure[4*3*i + 3] = 0.0f; drawing_structure[4*3*i + 4] = 1.0f; drawing_structure[4*3*i + 5] = cut;
				drawing_structure[4*3*i + 6] = 1.0f; drawing_structure[4*3*i + 7] = 1.0f; drawing_structure[4*3*i + 8] = cut;
				drawing_structure[4*3*i + 9] = 1.0f; drawing_structure[4*3*i + 10] = 0.0f; drawing_structure[4*3*i + 11] = cut;
			}
		}
		
		drawing_structure = new float[4*3*nodes.size()];
		drawing_recursion(0, new float[]{0.0f, 1.0f, 0.0f, 1.0f, 0.0f, 1.0f});
		
		return drawing_structure;
	}
	
	/**
	 * Recursion for generating the drawing structure.
	 * 
	 * @param p_index
	 * @param interval
	 */
	private void drawing_recursion(int p_index, float[] interval) {
		int d = nodes.get(p_index).dimension;
		float cut = (nodes.get(p_index).cut + 0.5f) / vol_size[d];// 100.0f; //
		if (nodes.get(p_index).cut == 0.0) cut = - 10000; //nice hack to fix it :P (moves the invalid ones far away...)
		if (d == dimension-1) cut /= (float) Short.MAX_VALUE /vol_size[d];// 100.0f; //
		
		if (d == 0) {
			drawing_structure[4*3*p_index + 0] = cut; 
			drawing_structure[4*3*p_index + 1] = interval[2]; 
			drawing_structure[4*3*p_index + 2] = interval[4];
			
			drawing_structure[4*3*p_index + 3] = cut; 
			drawing_structure[4*3*p_index + 4] = interval[2]; 
			drawing_structure[4*3*p_index + 5] = interval[5];
			
			drawing_structure[4*3*p_index + 6] = cut; 
			drawing_structure[4*3*p_index + 7] = interval[3]; 
			drawing_structure[4*3*p_index + 8] = interval[5];
			
			drawing_structure[4*3*p_index + 9] = cut;
			drawing_structure[4*3*p_index + 10] = interval[3]; 
			drawing_structure[4*3*p_index + 11] = interval[4];
			
			if (nodes.get(p_index).leftChild != -1) {
				//System.out.println("D0 drawing left " + p_index);
				drawing_recursion(nodes.get(p_index).leftChild, 
					new float[]{	interval[0], cut, 
									interval[2], interval[3], 
									interval[4], interval[5]});
			}
			if (nodes.get(p_index).rightChild != -1) {
				//System.out.println("D0 drawing right " + p_index);
				drawing_recursion(nodes.get(p_index).rightChild, 
					new float[]{	cut, interval[1], 
									interval[2], interval[3], 
									interval[4], interval[5]});
			}
		}
		
		if (d == 1) {
			drawing_structure[4*3*p_index + 0] = interval[0]; 
			drawing_structure[4*3*p_index + 1] = cut; 
			drawing_structure[4*3*p_index + 2] = interval[4];
			
			drawing_structure[4*3*p_index + 3] = interval[0]; 
			drawing_structure[4*3*p_index + 4] = cut; 
			drawing_structure[4*3*p_index + 5] = interval[5];
			
			drawing_structure[4*3*p_index + 6] = interval[1]; 
			drawing_structure[4*3*p_index + 7] = cut; 
			drawing_structure[4*3*p_index + 8] = interval[5];
			
			drawing_structure[4*3*p_index + 9] = interval[1]; 
			drawing_structure[4*3*p_index + 10] = cut; 
			drawing_structure[4*3*p_index + 11] = interval[4];
			
			if (nodes.get(p_index).leftChild != -1) {
				//System.out.println("D1 drawing left " + p_index);
				drawing_recursion(nodes.get(p_index).leftChild, 
					new float[]{	interval[0], interval[1], 
									interval[2], cut, 
									interval[4], interval[5]});
			}
			if (nodes.get(p_index).rightChild != -1) {
				//System.out.println("D1 drawing right " + p_index);
				drawing_recursion(nodes.get(p_index).rightChild, 
					new float[]{	interval[0], interval[1], 
									cut, interval[3], 
									interval[4], interval[5]});
			}
		}
		
		if (d == 2) {
			drawing_structure[4*3*p_index + 0] = interval[0]; 
			drawing_structure[4*3*p_index + 1] = interval[2]; 
			drawing_structure[4*3*p_index + 2] = cut;
			
			drawing_structure[4*3*p_index + 3] = interval[0]; 
			drawing_structure[4*3*p_index + 4] = interval[3]; 
			drawing_structure[4*3*p_index + 5] = cut;
			
			drawing_structure[4*3*p_index + 6] = interval[1]; 
			drawing_structure[4*3*p_index + 7] = interval[3]; 
			drawing_structure[4*3*p_index + 8] = cut;
			
			drawing_structure[4*3*p_index + 9] = interval[1]; 
			drawing_structure[4*3*p_index + 10] = interval[2]; 
			drawing_structure[4*3*p_index + 11] = cut;
			
			if (nodes.get(p_index).leftChild != -1) {
				//System.out.println("D2 drawing left " + p_index);
				drawing_recursion(nodes.get(p_index).leftChild, 
					new float[]{	interval[0], interval[1], 
									interval[2], interval[3], 
									interval[4], cut});
			}
			if (nodes.get(p_index).rightChild != -1) {
				//System.out.println("D2 drawing right " + p_index);
				drawing_recursion(nodes.get(p_index).rightChild, 
					new float[]{	interval[0], interval[1], 
									interval[2], interval[3], 
									cut, interval[5]});
			}
		}
		
	}
	
	/**
	 * Returns uv-coordinates since they are required in the current rendering process.
	 * 
	 * @return float[]
	 */
	public float[] get_drawingStructureUV() {
		
		float[] out = new float[4*2*nodes.size()];
		
		for (int i=0; i<out.length; i++) out[i] = 0.5f;
		
		return out;
		
	}
	
	public float[] get_drawingStructureColor() {
		
		float[] out = new float[4*3*nodes.size()];
		
		for (int i=0; i<out.length; i++) out[i] = 0.5f;
		
		return out;
		
	}
	
	/**
	 * Returns a position list of all leaf nodes.
	 * 
	 * @return ArrayList<GKDLeafNode>
	 */
	private ArrayList<GKDLeafNode> getLeavesList() {
		ArrayList<GKDLeafNode> leaves = new ArrayList<GKDLeafNode>();
		for (int i=0; i<nodes.size(); i++) {
			if (GKDLeafNode.class.isInstance(nodes.get(i))) {
				leaves.add((GKDLeafNode)nodes.get(i));
			}
		}
		return leaves;
	}
	
	/**
	 * Returns a position array of all leaf nodes.
	 * 
	 * Need to be scaled.. since they are in the array!
	 * 
	 * This is correct: Centroid of the bounding box, not centroid of the leaf node!
	 * 
	 * @return float[]
	 */
	public float[] get_Leaves() {
		
		ArrayList<GKDLeafNode> leaves = new ArrayList<GKDLeafNode>();
		
		for (int i=0; i<nodes.size(); i++) {
			if (GKDLeafNode.class.isInstance(nodes.get(i))) {
				leaves.add((GKDLeafNode)nodes.get(i));
			}
		}
		
		float[] out = new float[dimension * leaves.size()];
		
		for (int i=0; i<leaves.size(); i++) {
			float[] curr = leaves.get(i).pos;
			float value = leaves.get(i).value;
			
			//System.out.println(curr[0] + "/" + curr[1] + "/" + curr[2]);
			
			for (int d=0; d<dimension-1; d++) {
				// + 0.5f why? 
				out[dimension * i + d] = (curr[d] + 0.5f) / vol_size[d];// / 100.0f;
			}
			out[dimension * i + dimension-1] = value / (octree.getLimit() / 10.0f);
			//out[dimension * i + dimension-1] = (curr[dimension-1]) * vol_size[dimension-1] * (vol_size[dimension-1] / 8.0f);//curr[dimension-1] / octree.getLimit();// / (float) Short.MAX_VALUE * 1.0f;
		}
		
		return out;
	}
	
	public float[] get_LeavesColor() {
		
		ArrayList<GKDLeafNode> leaves = new ArrayList<GKDLeafNode>();
		
		for (int i=0; i<nodes.size(); i++) {
			if (GKDLeafNode.class.isInstance(nodes.get(i))) {
				leaves.add((GKDLeafNode)nodes.get(i));
			}
		}
		
		float[] out = new float[3 * leaves.size()];
		
		for (int i=0; i<leaves.size(); i++) {
			float[] curr = leaves.get(i).pos;
			
			//System.out.println(curr[0] + "/" + curr[1] + "/" + curr[2]);
			
			out[dimension * i + 0] = curr[dimension-1] / (octree.getLimit());
			for (int d=1; d<3; d++) {
				out[dimension * i + d] = 0.0f;//curr[dimension-1] / vol_size[dimension-1] / (float) Short.MAX_VALUE;//(float) Short.MAX_VALUE * 1.0f;
			}
			//out[dimension * i + dimension-1] = curr[dimension-1] / (float) Short.MAX_VALUE * 1.0f;
		}
		
		return out;
	}
	
}
