package io;

import java.util.ArrayList;

import org.lwjgl.input.Mouse;

import rendertools.RayCast;
import world.CameraFPS;
import world.SceneObject;
import world.VertexData;

/**
 *
 * Key mapping defines what happens if certain key states are set.
 * 
 * This still has heavy problems due to synchronization. An event may not be detected
 * the further down the handling of events is the larger the chance that an event
 * is correctly detected. (this makes a different for non polling requests) - FIX THIS
 * 
 * @author DarkMatt
 *
 */

public class KeyMapping {

	private float mouseSensitivity = 0.001f;
	private float dx, dy;
	
	private float[] debug_dir = new float[]{ 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -1.0f };
	private float[] selection_position = new float[]{0.0f, 0.0f, -3.0f};
	
	private int active_object;
	private boolean sigma_s_mode, sigma_r_mode;
	private String typing_s_string, typing_r_string;
	private boolean has_period;
	
	public KeyMapping() {
		Mouse.setGrabbed(true);
		sigma_s_mode = true;
		sigma_r_mode = false;
		typing_s_string = "0.05";
		typing_r_string = "0.4";
		has_period = true;
	}
	
	public void apply_key(
			KeyStates s, CameraFPS c, RayCast rc, 
			VertexData vd, ArrayList<SceneObject> so, 
			UiFileList ufl, 
			float dt) {
		
		//other states
		if (s.active_loading) ufl.setLoadingBar();
		
		//System.out.println(dt);
		
		if (s.move_fwd) 	c.translate(0.0f, 0.0f, dt * -3.5f);
		if (s.move_bwd) 	c.translate(0.0f, 0.0f, dt * +3.5f);
		if (s.move_left) 	c.translate(dt * -3.5f, 0.0f, 0.0f);
		if (s.move_right) 	c.translate(dt * +3.5f, 0.0f, 0.0f);
		if (s.move_up) 		c.translate(0.0f, dt * +3.5f, 0.0f);
		if (s.move_down) 	c.translate(0.0f, dt * -3.5f, 0.0f);
		
		if (s.rotate_left) 	c.rotY(dt * 1.3f);
		if (s.rotate_right) c.rotY(dt * -1.3f);
		
		//MOUSE
		dx = Mouse.getDX();
        dy = Mouse.getDY();
        
		c.rotX( dy * mouseSensitivity);
		c.rotY(-dx * mouseSensitivity);
		//MOUSE END
		
		if (s.action) {
			
			//TODO: vis HACK
			
			float[] direction = new float[]{
					c.get_position()[0], 
					c.get_position()[1], 
					c.get_position()[2],
					c.get_direction()[0],
					c.get_direction()[1],
					c.get_direction()[2]};
			
			debug_dir = direction;
			
			//System.out.println("dir: " + direction[3] + "/" + direction[4] + "/" + direction[5]);
			//System.out.println("pos: " + c.get_position()[0]  + "/" + c.get_position()[1] + "/" + c.get_position()[2]);
			
			active_object = rc.cast_ray(c.get_position(), c.get_direction());
			
			//System.out.println("pressed action: " + active_object);
			
			/*if (active_object == 4) {
				//so.get(index).position[0] += 1.0f;
				so.get(active_object).translate(0.0f, 0.0f, 1.0f);
				for (int i=0;i<so.size(); i++) {
					so.get(active_object).apply_object_transformation(vd);
				}
			}
			
			if (active_object == 5) {
				//so.get(index).position[0] += 1.0f;
				so.get(active_object).translate(1.0f, 0.0f, 0.0f);
				for (int i=0;i<so.size(); i++) {
					so.get(active_object).apply_object_transformation(vd);
				}
			}*/
			
		}
		
		if (s.actionactive) {
			if (active_object == 4) {
				
				float ny = 2.0f * dy * mouseSensitivity;
				float nz = 2.0f * dx * mouseSensitivity;
				
				if (selection_position[1] + ny - 0.6 > 0.0f || selection_position[1] + ny + 0.6025 < 0.0f) {
					ny = 0.0f;
				}
				if (selection_position[2] + nz - 0.6 > -3.0f || selection_position[2] + nz + 0.6025 < -3.0f) {
					nz = 0.0f;
				}
				
				//so.get(index).position[0] += 1.0f;
				so.get(4).translate(0.0f, ny, nz);
				so.get(5).translate(0.0f, ny, 0.0f);
				so.get(6).translate(0.0f, ny, nz);
				//so.get(active_object).translate(0.0f, 0.0f, 0.001f);
				
				so.get(4).apply_object_transformation(vd);
				so.get(5).apply_object_transformation(vd);
				
				selection_position[1] += ny;
				selection_position[2] += nz;
			}
			
			if (active_object == 5) {
				
				float nx = 2.0f * dx * mouseSensitivity;
				float ny = 2.0f * dy * mouseSensitivity;
				
				if (selection_position[0] + nx - 0.6 > 0.0f || selection_position[0] + nx + 0.6025 < 0.0f) {
					nx = 0.0f;
				}
				if (selection_position[1] + ny - 0.6 > 0.0f || selection_position[1] + ny + 0.6025 < 0.0f) {
					ny = 0.0f;
				}
				
				//so.get(index).position[0] += 1.0f;
				so.get(4).translate(0.0f, ny, 0.0f);
				so.get(5).translate(nx, ny, 0.0f);
				so.get(6).translate(nx, ny, 0.0f);
				
				so.get(4).apply_object_transformation(vd);
				so.get(5).apply_object_transformation(vd);
				
				selection_position[0] += nx;
				selection_position[1] += ny;
			}
		}
		
		//only works down here...
		if (s.arrow_up) 	ufl.move(true);//c.rotX(dt * 1.3f);
		if (s.arrow_down) 	ufl.move(false);//c.rotX(dt * -1.3f);
		if (s.arrow_left || s.arrow_right) swapSigmaMode();
		if (s.enter) 		{
			System.out.println("Enter pressed.");
			ufl.unblur();
		}
		if (s.v_key) {
			System.out.println("V pressed.");
			ufl.swapMode();
		}
		if (s.b_key) {
			System.out.println("B pressed.");
			ufl.blur();
		}
		if (sigma_s_mode || sigma_r_mode) {
			String typing_string = "";
			if (sigma_s_mode) {
				typing_string = typing_s_string;
			} else {
				typing_string = typing_r_string;
			}
			for (int i=0; i<s.digits.length; i++) {
				if (s.digits[i]) {
					//System.out.println("Pressed: " + (i+1)%10);
					typing_string += (i+1)%10;
				}
			}
			
			if (s.back_key) {

				int length = typing_string.length();
				if (length > 0) {
					if (typing_string.charAt(length-1) == '.') has_period = false;
					typing_string = typing_string.substring(0, (length-1));
				}
			}
			if (s.dot_key) {
				if (!has_period) {
					typing_string += ".";
					has_period = true;
				}
			}
			if (sigma_s_mode) {
				typing_s_string = typing_string;
			} else {
				typing_r_string = typing_string;
			}
			
		} else { //default
			
		}
		
		//TODO: Reset key-states
		s.reset();
		
	}
	
	public String getTypingSString() {
		return typing_s_string;
	}
	
	public String getTypingRString() {
		return typing_r_string;
	}
	
	public void swapSigmaMode() {
		sigma_s_mode = !sigma_s_mode;
		sigma_r_mode = !sigma_r_mode;
	}
	
	public boolean sActive() {
		return sigma_s_mode;
	}
	
	public float[] get_debug_dir() {
		return debug_dir;
	}
	
	public float[] get_seclection_position() {
		return selection_position;
	}
	
}
