package io;

import gkdtree.FileNode;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;

import org.lwjgl.opengl.GL11;

import rendertools.TextRenderer;
import rendertools.Textures;

public class UiFileList {

	private ArrayList<String> filenames;
	private int current_selection = 0;
	private int text_start_index;
	
	private boolean loadingbar = false;
	private int draw_mode = 0;
	private boolean blur = false;
	
	public UiFileList(Textures t) {
		blur = false;
		filenames = new ArrayList<String>();
		getFiles(t);
	}
	
	private void getFiles(Textures t) {
		
		URI uri = null;
		try {
			uri = new URI("file://"+UiFileList.class.getProtectionDomain().getCodeSource().getLocation().getPath());
		} catch (URISyntaxException e) {
			e.printStackTrace();
		}
		File absolute_path = new File(new File(uri).getParent());
		
		String path = "/data/"; 
		String files;
		File folder = new File(absolute_path + path);
		File[] listOfFiles = folder.listFiles();  
		for (int i = 0; i < listOfFiles.length; i++) {
			if (listOfFiles[i].isFile()) {
				files = listOfFiles[i].getName();
				if (files.endsWith(".png") || files.endsWith(".PNG")) {
					//System.out.println(files);
					filenames.add(files);
				}
			}
		}
		
		for(int i=0; i<filenames.size()-1; i++) System.out.print(filenames.get(i)+", ");
		System.out.println(filenames.get(filenames.size()-1));
		
		text_start_index = -1;
		for (int i=0; i<filenames.size(); i++) {
			if (text_start_index == -1)
				text_start_index = t.add_Texture_Full_Name(filenames.get(i));
			else
				t.add_Texture_Full_Name(filenames.get(i));
		}
		
	}
	
	public void move(boolean up) {
		if (up) {
			current_selection --;
			if (current_selection == -1) current_selection = filenames.size()-1;
		} else {
			current_selection ++;
			if (current_selection == filenames.size()) current_selection = 0;
		}
	}
	
	public void setLoadingBar() {
		loadingbar = true;
	}
	
	public void swapMode() {
		draw_mode ++;
		if (draw_mode == 2) draw_mode = 0;
	}
	
	public float getMode() {
		if (draw_mode == 0) return 0.0f;
		else return 1.0f;
	}
	
	public void blur() {
		blur = true;
	}
	
	public void unblur() {
		blur = false;
	}
	
	public boolean isBlurred() {
		return blur;
	}
	
	public String getFilename() {
		return filenames.get(current_selection);
	}
	
	public int getCurrentImgeIndex() {
		return current_selection + text_start_index - 1;
	}
	
	public int[] getSize(Textures t) {
		if (text_start_index != -1) {
			int text_ind = current_selection + text_start_index - 1;
			return t.getSize(text_ind);
		} else {
			return new int[]{0,0};
		}
	}
	
	public int getFilenameIndex(String name) {
		int out = 0;
		
		for (int i=0; i<filenames.size(); i++) {
			if (filenames.get(i).equals(name)) out = i;
		}
		
		return out;
	}
	
	public void renderList(TextRenderer textrenderer, Textures t) {
		
	    if (loadingbar) {
	    	textrenderer.draw_string("Loading file...", -0.07f, 0.9f, 0.3f);
	    	loadingbar = false;
	    }
		
		float o = 0.45f;
		textrenderer.draw_string("Arrow-keys to select:", -1.23f, o, 0.27f);
		GL11.glColor3f(0.2f, 0.2f, 0.2f);
		for (int i=0; i<filenames.size(); i++) {
			if (i == current_selection) GL11.glColor3f(0.99f, 0.99f, 0.9f);
			if (i-1 == current_selection) GL11.glColor3f(0.2f, 0.2f, 0.2f);
			textrenderer.draw_string(filenames.get(i), -1.2f, o - 0.07f -i*0.07f, 0.4f);
		}
		
		if (text_start_index != -1) {
			int text_ind = current_selection + text_start_index - 1;
			
			GL11.glColor3f(1.0f, 1.0f, 1.0f);
			
			if (blur) {
				textrenderer.draw_string("blurred"
						, -0.0f, -0.8f, 0.29f);
			}
			
			textrenderer.draw_string(
					getSize(t)[0] +"   x "+ getSize(t)[1]
					, -1.2f, 0.515f, 0.2f);
			
			GL11.glEnable(GL11.GL_TEXTURE_2D);
			
			t.bind_texture(text_ind);
			
			float uv[] = t.getOffset(text_ind);
			GL11.glBegin(GL11.GL_QUADS);
				GL11.glTexCoord2f(0,  0);
		        GL11.glVertex2f(-1.20f,  0.95f);
		        GL11.glTexCoord2f(uv[0],  0);
		        GL11.glVertex2f(-0.80f,  0.95f);
		        GL11.glTexCoord2f(uv[0],  uv[1]);
		        GL11.glVertex2f(-0.80f,  0.55f);
		        GL11.glTexCoord2f(0,  uv[1]);
		        GL11.glVertex2f(-1.20f,	 0.55f);
		    GL11.glEnd();
			
		    GL11.glDisable(GL11.GL_TEXTURE_2D);
		    
		}
	}
	
}
