package main;
import java.awt.Dimension;
import java.awt.Toolkit;

import io.Input;
import io.KeyStates;
import world.Scene;

/**
 * 
 * @author DarkMatt
 *
 * Handles very basic multi-threading and init/destroy order.
 * 
 */

public class RuntimeManager {
	
	private static Input i;
	/**
	 * Only available in this class.
	 */
	private static Scene s;
	private static Window w;
	
	public boolean run() {
		
		/*
		 * Comments:
		 *  *) w-thread currently runs independent of this update loop
		 */
		
		
		//input
		i.update_Key_States();
		i.draw_User_Interface();
		
		//think (game logic per object)
			
		//move
		//currently in w-thread
			
		//messaging (inter object communication)
			
		//camera
		//currently in w-thread
		
		//render
		KeyStates k = i.get_Input_State(); //source of trouble? (used to be 2 calls)
		
		//Update geometry
		if ((k.enter || k.b_key) && s.was_Load) {
			if (k.enter) s.activateLoading(w.getFilterParams(),false);
			if (k.b_key) s.activateLoading(w.getFilterParams(),true);
			
			k.active_loading = true;
		} else if (!s.isLoading()) {
			//done with loading...
			k.active_loading = false;
		}
		
		w.apply_Keys(k);
		
		return !k.shut_down;
		
	}
	
	public void start() {
		
		if (Entrance.DEBUG) System.out.println("RuntimeManager: Starting...");
		
		// Get the current screen size
		Dimension screen_size = Toolkit.getDefaultToolkit().getScreenSize();
		
		i = new Input();
		
		//start window thread
		w = new Window("MetaSpace", true, screen_size.width, screen_size.height);
		w.start(); //creates a new window
		
		s = new Scene(screen_size.width, screen_size.height);
		s.start();
		/*try {
			s.join(); //necessary?
		} catch (InterruptedException e) {
			e.printStackTrace();
		}*/
		while (!s.was_Load) try { Thread.sleep(100); } catch (InterruptedException e) { e.printStackTrace(); }
		
		//Needs to wait also until the init in window is done!
		w.set_vertex_data(s.get_vertex_data(), s.get_scene_object(), s.get_current_camera());
		w.init_raycasting();
		w.interrupt();
				
		//wait until window was created
		while (!w.is_Created()) try { Thread.sleep(100); } catch (InterruptedException e) { e.printStackTrace(); }
		
		
		//additional single threaded loading:
		//i.init_User_Interface(w.get_UI_hook());
		
		
		if (Entrance.DEBUG) System.out.println("RuntimeManager: Started.");
		
	}
	
	public void stop() {
		
		if (Entrance.DEBUG) System.out.println("RuntimeManager: Shutting down...");
		
		s.interrupt();
		//Closes the window, should be the last.
		w.shut_down();
		
		if (Entrance.DEBUG) System.out.println("RuntimeManager: Shut down.");
		
	}
	
}
