package main;
import gkdtree.FilterParameters;
import io.KeyMapping;
import io.KeyStates;

import java.util.ArrayList;

import org.lwjgl.LWJGLException;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.GL11;

import rendertools.RayCast;
import rendertools.RendererMain;
import rendertools.Timer;
import world.CameraFPS;
import world.SceneObject;
import world.VertexData;

/**
 *
 * Window implementation. Window generation and entry point for rendering 
 * and other window specific functionality.
 * 
 * @author DarkMatt
 *
 */

public class Window extends Thread {

	private boolean is_Created = false;
	
	private String window_name;
	private boolean fullscreen;
	private boolean running;
	private float aspect;
	private int px, py;
	private float sx, sy;
	private boolean loading;
	private RendererMain renderer;
	
	private boolean done_loading;
	private Timer timer;
	private KeyStates key_states;
	private KeyMapping key_mapping;
	private VertexData vd;
	private ArrayList<SceneObject> so;
	private CameraFPS c;
	private RayCast rc;
	
	private boolean active_raycasting = false;
	
	//private GUI gui;
	
	private float[] debug_dir = new float[]{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f};
	
	public Window(String name, boolean fullscreen, int sx, int sy) {
		this.window_name = name;
		this.fullscreen = fullscreen;
		this.px = sx;
		this.py = sy;
		loading = true;
		done_loading = false;
		timer = new Timer();
		key_states = new KeyStates();
		key_mapping = new KeyMapping();
	}
	
	private void init() {
		/*
		 * ad aspect:
		 * parameterization of the screen in -1 to 1 along the minimal direction
		 */
		aspect = (float)px/(float)py;
		if (aspect > 1.0) {
			this.sx = aspect;
			this.sy = 1.0f;
		} else {
			this.sx = 1.0f;
			this.sy = 1.0f/aspect;
		}
		
		running = true;
		
		if (Entrance.DEBUG) System.out.println("1: Open window");
		
		//takes a while... (attention)
		Display.setTitle(window_name);
		try {
			Display.setFullscreen(fullscreen);
			//Display.setDisplayMode(new DisplayMode(px, py));
			Display.create();
		} catch (LWJGLException e) {
			e.printStackTrace();
		}
		
		//create_UI();
		
		
		GL11.glEnable(GL11.GL_DEPTH_TEST);
		GL11.glEnable(GL11.GL_TEXTURE_2D);
		
		if (Entrance.DEBUG) System.out.println("2: Done window");
		
		if (Entrance.DEBUG) System.out.println("3: Creating renderer");
		renderer = new RendererMain(sx, sy);
		
		if (Entrance.DEBUG) System.out.println("4: Init done.");
		
		is_Created = true;
	}
	
	public void init_raycasting() {
		rc = new RayCast(vd.getVertices(), vd.getVerticesSizeForRaycast(), 4);
		active_raycasting = true;
	}
	
	/*
	private void create_UI() {
		try {
			ui_hook = new LWJGLRenderer();
		} catch (LWJGLException e) {
			e.printStackTrace();
		}
		
		gui = new GUI(new UILayer(), ui_hook);
		
		URI uri = null;
		try {
			uri = new URI("file://"+Input.class.getProtectionDomain().getCodeSource().getLocation().getPath());
		} catch (URISyntaxException e1) {
			throw new RuntimeException("URI fault");
		}
        File f = new File(uri);
		
		ThemeManager theme;
		try {
			theme = ThemeManager.createThemeManager(
					new URL(("file:///"+f.getParent()+"/content/ui/ui.xml")), ui_hook);
			System.out.println("Could not find theme?");
			gui.applyTheme(theme);
			System.out.println("Could not find theme!");
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}*/
	
	//seems to be in a different thread...
	public void set_vertex_data(VertexData vd, ArrayList<SceneObject> so, CameraFPS c) {
		
		this.vd = vd;
		this.so = so;
		this.c = c;
		
		/*while (!done_loading) {
			try {
				sleep(100);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		
		System.out.println("Uploading vertex data...");
		
		//init here removes the exception?
		renderer.upload_geometry(vd);*/
	}
	
	//TODO: now asynchronous
	public void apply_Keys(KeyStates key_states) {
		this.key_states = key_states;
	}
	
	public boolean is_Created() {
		return is_Created;
	}
	
	public void run() {
		
		init();
		
		if (Entrance.DEBUG) System.out.println("Starting the render loop");
		
		//game loop
		while(running) { //

			timer.startMeasure();
			
			if (active_raycasting) { //TODO: needs to start after scene...
				//input
				key_mapping.apply_key(key_states, c, rc, vd, so, renderer.getUiFileList(), timer.dt());
				//loadFilterInput()
				if (key_states.actionactive) {
					renderer.set_Selection_Pos(key_mapping.get_seclection_position());
					renderer.update_geometry(vd, 48, 8);
				}
				if (vd.was_modified()) {
					System.out.println("Out of synch: " + vd.point_data[0] +"/"+ vd.point_data[1]);
					renderer.update_geometry(vd, vd.point_data[0], vd.point_data[1] + vd.point_data[0] );
				}
			}
			
			debug_dir = key_mapping.get_debug_dir();
			
			if (active_raycasting) {
				String ray_hit = String.valueOf(rc.cast_ray(c.get_position(), c.get_direction()));
				renderer.set_dbug_text(ray_hit);
			}
			
			//do nothing
			
			if (loading) {
				renderer.draw_loading();
			} else {
				
				if (!done_loading) {
					renderer.upload_geometry(vd, so, c);
					
					//System.out.println("outer loop: " + vd.getVertices()[4*3+1]);
					
					done_loading = true;
				}
				
				renderer.draw_scene();
				
				renderer.draw_debug_dir(debug_dir);
				
				renderer.draw_ui(timer.getFPS(), key_mapping.sActive(), key_mapping.getTypingSString(), key_mapping.getTypingRString());
				
				//wgui.update();
				
			}
			
			if (Thread.interrupted()) {
				if (Entrance.DEBUG) System.out.println("Scene ready!");
				loading = false;
			}
			
			Display.update();
			
			//for (int i=0; i<100000; i++) System.out.print("");
			
			timer.endMeasure();
			
		}
		
		Display.destroy();
		
	}
	
	public FilterParameters getFilterParams() {
		return renderer.getFilterParams();
	}
	
	public void shut_down() {
		
		if (Entrance.DEBUG) System.out.println("Window: Shutting down...");
		
		running = false;
		
	}
	
}
