package rendertools;

/**
 *
 * Ray-Casting implementation.
 * 
 * @author DarkMatt
 *
 */

public class RayCast {

	//Fast MinimumStorage RayTriangle Intersection.pdf
	//excellent: http://paulbourke.net/geometry/planeline/
	
	private static final float EPSILON = 0.00001f;
	
	private float[] vertex_data;
	private float[] aabb; //AxisAlignedBoundingBox: -x, x, -y, y, -z, z
	private int num_faces;
	
	private boolean quads = false;
	
	/*
	 * AABB NOT USED YET!
	 * 
	 */
	
	public RayCast(float[] vertex_data, int length, int num_per_face) {
		
		//System.out.println("-------RAYCAST-------");
		//System.out.println("length: " + length + ", num_per_face: " + num_per_face);
		
		this.vertex_data = vertex_data;
		this.num_faces = length/num_per_face;
		
		int per_primitive = 6;
		
		if (num_per_face == 4) {
			quads = true;
			aabb = new float[6*2*num_faces];
			per_primitive = 12;
		} else {
			aabb = new float[6*num_faces];
		}
		
		for (int i=0; i<num_faces; i++) {
			//for triangles
			float[] minmax = Extremum.getMinMax(vertex_data, num_per_face*i, 9, false);
			aabb[per_primitive*i+0] = minmax[0];
			aabb[per_primitive*i+1] = minmax[1];
			aabb[per_primitive*i+2] = minmax[2];
			aabb[per_primitive*i+3] = minmax[3];
			aabb[per_primitive*i+4] = minmax[4];
			aabb[per_primitive*i+5] = minmax[5];
			if (quads) {
				minmax = Extremum.getMinMax(vertex_data, num_per_face*i, 9, true);
				aabb[per_primitive*i+6] = minmax[0];
				aabb[per_primitive*i+7] = minmax[1];
				aabb[per_primitive*i+8] = minmax[2];
				aabb[per_primitive*i+9] = minmax[3];
				aabb[per_primitive*i+10] = minmax[4];
				aabb[per_primitive*i+11] = minmax[5];
			}
		}
	}
	
	//"brute force": Fast MinimumStorage RayTriangle Intersection.pdf
	private float intersect(float[] start, float[] ds, int i, boolean quad_2nd) {
		
		//special case for quads
		int offset = 0;
		if (quad_2nd) offset = 6;
		
		float[] vert0 = new float[]{
				vertex_data[offset + 3*i+0], 
				vertex_data[offset + 3*i+1], 
				vertex_data[offset + 3*i+2]};
		
		float[] edge1 = VecMath.sub(
				new float[]{
						vertex_data[offset + 3*i+3], 
						vertex_data[offset + 3*i+4], 
						vertex_data[offset + 3*i+5]}, 	//VERTEX1
				vert0);												//VERTEX0
		
		//special case for quads
		if (quad_2nd) offset = -6;
		
		float[] edge2 = VecMath.sub(
				new float[]{
						vertex_data[offset + 3*i+6], 
						vertex_data[offset + 3*i+7], 
						vertex_data[offset + 3*i+8]}, 	//VERTEX2
				vert0);									//VERTEX0
		
		float[] pvec = VecMath.cross(ds, edge2);
		
		float det = VecMath.dot(edge1, pvec);
		
		if (det > -EPSILON && det < EPSILON) return Float.MAX_VALUE;
		
		float inv_det = 1.0f/det;
		
		float[] tvec = VecMath.sub(start, vert0);
		
		float u = VecMath.dot(tvec, pvec)*inv_det;
		
		if (u < 0.0f || u > 1.0f) return Float.MAX_VALUE;
		
		float[] qvec = VecMath.cross(tvec, edge1);
		
		float v = VecMath.dot(ds, qvec)*inv_det;
		
		if (v < 0.0f || u+v > 1.0f) return Float.MAX_VALUE;
		
		return VecMath.dot(edge2,qvec)*inv_det;
	}
	
	public int cast_ray(float[] start, float[] ds) {
		
		float distance = Float.MAX_VALUE;
		int index = -1;
		
		int num_per_face = 3;
		if (quads) num_per_face = 4;
		
		for (int i=0; i<num_faces; i++) {
			float d = intersect(start, ds, num_per_face*i, false);
			if (d<distance && d > 0.0f) { //TODO: hmmm: projection matrix starts from 1.0f
				//System.out.println(i + ": " + d);
				distance = d;
				index = i;
			}
			if (quads) {
				d = intersect(start, ds, num_per_face*i, true);
				if (d<distance && d > 0.0f) { //TODO: hmmm: projection matrix starts from 1.0f
					//System.out.println(i + ": " + d);
					distance = d;
					index = i;
				}
			}
		}
		
		return index;
	}
	
}
