package rendertools;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;

import main.Entrance;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL21;
import org.lwjgl.opengl.GL32;
import org.lwjgl.opengl.GLContext;
import org.lwjgl.opengl.ARBShaderObjects;
import org.lwjgl.opengl.ARBVertexShader;
import org.lwjgl.opengl.ARBFragmentShader;

//TODO: error checking

/**
 *
 * Basic Shader implementation. (To be extended for multiple shader combinations.)
 * 
 * @author DarkMatt
 *
 */

public class Shader {

	private int vertex_shader = 0;
	private int fragment_shader = 0;
	private int geometry_shader = 0;
	private int program = 0;
	private FloatBuffer matrix = BufferUtils.createFloatBuffer(16);
	
	private boolean disable_error_msg = true;
	private boolean debug = false;
	
	//is filled with the UNIFORM location connections
	private ArrayList<Integer> value_ids = new ArrayList<Integer>();
	
	private boolean pre_opengl20 = false;
	private boolean geom_shader = true;
	
	Shader(String path) {
		
		
		
		String vert_path = path + ".vert";
		String geom_path = path + ".geom";
		String frag_path = path + ".frag";
		
		//CHECK version
		String version = GL11.glGetString(GL11.GL_VERSION);
		if (Integer.valueOf(version.substring(0,1)) < 2) {
			//System.out.println("version too low");
			pre_opengl20 = true;
			if (!GLContext.getCapabilities().GL_ARB_shader_objects) 
				throw new RuntimeException("Pre OpenGL 2.0 without GL_ARB_shader_objects support.");
			if (!GLContext.getCapabilities().GL_ARB_vertex_shader) 
				throw new RuntimeException("Pre OpenGL 2.0 without GL_ARB_vertex_shader support.");
			if (!GLContext.getCapabilities().GL_ARB_fragment_shader) 
				throw new RuntimeException("Pre OpenGL 2.0 without GL_ARB_vertex_shader support.");
		}
		
		
		String line = "";
		URI uri = null;
		try {
			uri = new URI("file://"+Shader.class.getProtectionDomain().getCodeSource().getLocation().getPath());
		} catch (URISyntaxException e1) {
			throw new RuntimeException("URI fault");
		}
		
        File f = new File(uri);
		BufferedReader reader;
		
		
		/** VERTEX SHADER */
		String vertex_shader_source = "";
		try {
			reader = new BufferedReader(new FileReader(f.getParent() + vert_path));
			while( (line=reader.readLine()) != null){
				vertex_shader_source += line + "\n";
			}
		} catch (Exception e) {
			throw new RuntimeException("Couldnt load shader! Was looking for: " + f.getParent() + vert_path +
					" \nReason: " + e);
		}
		if (!pre_opengl20) {
			vertex_shader = compile(GL20.GL_VERTEX_SHADER, vertex_shader_source);
		} else {
			vertex_shader = compilePre20(ARBVertexShader.GL_VERTEX_SHADER_ARB, vertex_shader_source);
		}
		if (vertex_shader == 0) System.out.println("Error Vertex Shader!");
		
		
		/** GEOMETRY SHADER */
		String geometry_shader_source = "";
		try {
			reader = new BufferedReader(new FileReader(f.getParent() + geom_path));
			while( (line=reader.readLine()) != null){
				geometry_shader_source += line + "\n";
			}
		} catch (Exception e) {
			if (Entrance.DEBUG) System.out.println("No geometry shader for: " + geom_path);
			geom_shader = false;
		}
		//TODO: version checking...
		if (geom_shader) {
			geometry_shader = compile(GL32.GL_GEOMETRY_SHADER, geometry_shader_source);
			if (geometry_shader == 0) System.out.println("Error Geometry Shader!");
		}
		
		
		/** FRAGMENT SHADER */
		String fragment_shader_source = "";
		try {
			reader = new BufferedReader(new FileReader(f.getParent() + frag_path));
			while( (line=reader.readLine()) != null){
				fragment_shader_source += line + "\n";
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		if (!pre_opengl20) {
			fragment_shader = compile(GL20.GL_FRAGMENT_SHADER, fragment_shader_source);
		} else {
			fragment_shader = compilePre20(ARBFragmentShader.GL_FRAGMENT_SHADER_ARB, fragment_shader_source);
		}
		if (fragment_shader == 0) System.out.println("Error Fragment Shader!");
		
		
		/** LINK */
		
		if (!pre_opengl20) {
			link(path);
		} else {
			linkPre20(path);
		}
		
		if (Entrance.DEBUG) System.out.println("Successfully loaded " + path + "!");
		
	}
	
	private int compilePre20(int type, String shader_source) {
		
		int shader = ARBShaderObjects.glCreateShaderObjectARB(type);
		if (shader == 0) {
			System.out.println("Impossible to create shader object!");
			return 0;
		}
		
		ARBShaderObjects.glShaderSourceARB(shader, shader_source);
		ARBShaderObjects.glCompileShaderARB(shader);
		
		return shader;
	}
	
	private int compile(int type, String shader_source) {
		
		int shader = GL20.glCreateShader(type);
		if (shader == 0) {
			System.out.println("Impossible to create shader object!");
			return 0;
		}
		
		GL20.glShaderSource(shader, shader_source);
		GL20.glCompileShader(shader);
		
		return shader;
	}
	
	private void linkPre20(String source) {
		
		program = ARBShaderObjects.glCreateProgramObjectARB();
		
		ARBShaderObjects.glAttachObjectARB(program, vertex_shader);
		ARBShaderObjects.glAttachObjectARB(program, fragment_shader);
		
		ARBShaderObjects.glLinkProgramARB(program);
		
		IntBuffer result = BufferUtils.createIntBuffer(1);
		ARBShaderObjects.glGetObjectParameterARB(program, ARBShaderObjects.GL_OBJECT_LINK_STATUS_ARB, result);
		
		//System.out.println(result.get(0));
		
		if(result.get(0) != GL11.GL_TRUE){
			throw new RuntimeException("Shader compile error! \n\n" + 
					"in: " + source + "\n\n" +
					ARBShaderObjects.glGetInfoLogARB(program, 5000));
		}
		
		System.out.println(ARBShaderObjects.glGetInfoLogARB(program, 5000));
		
	}
	
	private void link(String source) {
		
		program = GL20.glCreateProgram();
		
		GL20.glAttachShader(program, vertex_shader);
		if (geom_shader) GL20.glAttachShader(program, geometry_shader);
		GL20.glAttachShader(program, fragment_shader);
		
		GL20.glLinkProgram(program);
		
		IntBuffer result = BufferUtils.createIntBuffer(1);
		GL20.glGetProgram(program, GL20.GL_LINK_STATUS, result);
		
		//System.out.println(result.get(0));
		
		if(result.get(0) != GL11.GL_TRUE){
			throw new RuntimeException("Shader compile error! \n\n" + 
					"in: " + source + "\n\n" +
					GL20.glGetProgramInfoLog(program, 5000));
		}
		
		System.out.println(GL20.glGetProgramInfoLog(program, 5000));
		
	}
	
	public void bind() {
		if (!pre_opengl20) {
			GL20.glUseProgram(program);
		} else {
			ARBShaderObjects.glUseProgramObjectARB(program);
		}
	}
	
	public void unbind() {
		if (!pre_opengl20) {
			GL20.glUseProgram(0);
		} else {
			ARBShaderObjects.glUseProgramObjectARB(0);
		}
	}
	
	public int initValue1i(String name) {
		int location = 0;
		if (!pre_opengl20) {
			location = GL20.glGetUniformLocation(program, name);
		} else {
			location = ARBShaderObjects.glGetUniformLocationARB(program, name);
		}
		
		if (!disable_error_msg && location == -1) System.out.println("Error binding shader int.");
		
		value_ids.add(location);
		
		if (debug) System.out.println("Added: " + name + " at " + (value_ids.size()-1) + " location: " + location);
		
		return value_ids.size()-1;
	}
	
	public void setValue1i(int position_in_location_array, int value) {
		
		if (!pre_opengl20) {
			GL20.glUniform1i(value_ids.get(position_in_location_array), value);
		} else {
			ARBShaderObjects.glUniform1iARB(value_ids.get(position_in_location_array), value);
		}
		
	}
	
	public int initValue1f(String name) {
		int location = 0;
		if (!pre_opengl20) {
			location = GL20.glGetUniformLocation(program, name);
		} else {
			location = ARBShaderObjects.glGetUniformLocationARB(program, name);
		}
		
		if (!disable_error_msg && location == -1) System.out.println("Error binding shader float.");
		
		value_ids.add(location);
		
		if (debug) System.out.println("Added: " + name + " at " + (value_ids.size()-1) + " location: " + location);
		
		return value_ids.size()-1;
	}
	
	public void setValue1f(int position_in_location_array, float value) {
		
		if (!pre_opengl20) {
			GL20.glUniform1f(value_ids.get(position_in_location_array), value);
		} else {
			ARBShaderObjects.glUniform1fARB(value_ids.get(position_in_location_array), value);
		}
		
	}
	
	public int initValueVec3f(String name) {
		int location = 0;
		if (!pre_opengl20) {
			location = GL20.glGetUniformLocation(program, name);
		} else {
			location = ARBShaderObjects.glGetUniformLocationARB(program, name);
		}
		
		if (!disable_error_msg && location == -1) System.out.println("Error binding shader vec3.");
		
		value_ids.add(location);
		
		if (debug) System.out.println("Added: " + name + " at " + (value_ids.size()-1) + " location: " + location);
		
		return value_ids.size()-1;
	}
	
	public void setValueVec3f(int position_in_location_array, float x, float y, float z) {
		
		if (!pre_opengl20) {
			GL20.glUniform3f(value_ids.get(position_in_location_array), x, y, z);
		} else {
			ARBShaderObjects.glUniform3fARB(value_ids.get(position_in_location_array), x, y, z);
		}
		
	}
	
	public int initValueMat3f(String name) {
		int location = 0;
		if (!pre_opengl20) {
			location = GL20.glGetUniformLocation(program, name);
		} else {
			location = ARBShaderObjects.glGetUniformLocationARB(program, name);
		}
		
		if (!disable_error_msg && location == -1) System.out.println("Error binding shader matrix.");
		
		value_ids.add(location);
		
		if (debug) System.out.println("Added: " + name + " at " + (value_ids.size()-1) + " location: " + location);
		
		return value_ids.size()-1;
	}
	
	public void setValueMat3f(int position_in_location_array, float [] value) {
		
		matrix.clear();
		for(int i=0; i<9; i++) {
			matrix.put(value[i]);
		}
		matrix.flip();
		
		if (!pre_opengl20) {
			GL20.glUniformMatrix3(value_ids.get(position_in_location_array), true, matrix);
		} else {
			ARBShaderObjects.glUniformMatrix3ARB(value_ids.get(position_in_location_array), true, matrix);
		}
		
	}
	
	public int initValueMat4f(String name) {
		int location = 0;
		if (!pre_opengl20) {
			location = GL20.glGetUniformLocation(program, name);
		} else {
			location = ARBShaderObjects.glGetUniformLocationARB(program, name);
		}
		
		if (!disable_error_msg && location == -1) System.out.println("Error binding shader matrix.");
		
		value_ids.add(location);
		
		if (debug) System.out.println("Added: " + name + " at " + (value_ids.size()-1) + " location: " + location);
		
		return value_ids.size()-1;
	}
	
	public void setValueMat4x3f(int position_in_location_array, FloatBuffer value) {
		
		//matrix.clear();
		//for(int i=0; i<16; i++) {
		//	matrix.put(value[i]);
		//}
		//matrix.flip();
		
		if (!pre_opengl20) {
			GL21.glUniformMatrix4x3(value_ids.get(position_in_location_array), true, value);
		} else {
			throw new RuntimeException("Shader: 4x4 Matrix not implemented!");
			//TODO!
			//ARBShaderObjects.glUniformMatrix4ARB(value_ids.get(position_in_location_array), true, value);
		}
		
	}
	
	public void setValueMat4f(int position_in_location_array, FloatBuffer value) {
		
		if (!pre_opengl20) {
			GL20.glUniformMatrix4(value_ids.get(position_in_location_array), true, value);
		} else {
			ARBShaderObjects.glUniformMatrix4ARB(value_ids.get(position_in_location_array), true, value);
		}
		
	}
	
	/*
	public void setValueMat4f(int position_in_location_array, float [] value) {
		
		matrix.clear();
		for(int i=0; i<16; i++) {
			matrix.put(value[i]);
		}
		matrix.flip();
		
		if (!pre_opengl20) {
			GL20.glUniformMatrix4(value_ids.get(position_in_location_array), true, matrix);
		} else {
			ARBShaderObjects.glUniformMatrix4ARB(value_ids.get(position_in_location_array), true, matrix);
		}
		
	}*/
	
}
