package world;

import java.util.ArrayList;

/**
 *
 * Holds the raw vertex data. This is maintained on the main memory for collision detection, 
 * selection and more. It is the data-structure used for ray tracing and for draw calls 
 * on the GPU, which are uploaded as VBO and updates if necessary (at this point).
 * 
 * @author DarkMatt
 * 
 */

public class VertexData {

	/*
	 * Stores the raw vertex data
	 * 
	 * - drawing vertices (normally all)
	 * - collision vertices (subset)
	 * 
	 */
	
	public float[] vertices;
	private float[] textCoords;
	private float[] vertColors;
	public ArrayList<Integer> object_indices;
	public ArrayList<Boolean> for_raycast;
	
	private int current_size;
	private int raycast_size;
	private int max_size = 10000000; //TODO
	
	private boolean out_of_sync = false;
	
	/**
	 * To define start and end of the point data
	 */
	public int[] point_data;
	
	/*
	 * Annoying functions:
	 *  - add object
	 *  - modify object! (dynamic resize!)
	 */
	
	public VertexData() {
		current_size = 0;
		raycast_size = 0;
		vertices = new float[3*max_size];
		textCoords = new float[2*max_size];
		vertColors = new float[3*max_size];
		object_indices = new ArrayList<Integer>();
		for_raycast = new ArrayList<Boolean>();
		point_data = new int[2];
	}
	
	public void addObject(float[] v, float[] t, float[] c, int size, boolean ray_cast) {
		
		//System.out.println("added object: " + object_indices.size() + " at " + size);
		
		object_indices.add(v.length/3);
		for_raycast.add(ray_cast);
		
		if (current_size + size < max_size) {
			for (int i=0; i<size; i++) {
				vertices[3*current_size + 3*i] = v[3*i];
				vertices[3*current_size + 3*i+1] = v[3*i+1];
				vertices[3*current_size + 3*i+2] = v[3*i+2];
			}
			for (int i=0; i<size; i++) {
				textCoords[2*current_size + 2*i] = t[2*i];
				textCoords[2*current_size + 2*i+1] = t[2*i+1];
			}
			if (c != null) {
				for (int i=0; i<size; i++) {
					vertColors[3*current_size + 3*i] = c[3*i];
					vertColors[3*current_size + 3*i+1] = c[3*i+1];
					vertColors[3*current_size + 3*i+2] = c[3*i+2];
				}
			}
		}
		
		current_size += v.length/3;//size;
		
		if (ray_cast) raycast_size += v.length/3;//size;
		
		out_of_sync = true;
	}
	
	public void overwriteObject(float[] v, float[] t, float[] c, int start, int size, boolean ray_cast) {
		
		//System.out.println("added object: " + object_indices.size() + " at " + size);
		
		//start /= 3;
		//size /= 3;
		
		//object_indices.add(size);
		//for_raycast.add(ray_cast);
		
		//if (current_size + size < max_size) {
			for (int i=0; i<size; i++) {
				vertices[3*start + 3*i] = v[3*i];
				vertices[3*start + 3*i+1] = v[3*i+1];
				vertices[3*start + 3*i+2] = v[3*i+2];
			}
			for (int i=0; i<size; i++) {
				textCoords[2*start + 2*i] = t[2*i];
				textCoords[2*start + 2*i+1] = t[2*i+1];
			}
			if (c != null) {
				for (int i=0; i<size; i++) {
					vertColors[3*start + 3*i] = c[3*i];
					vertColors[3*start + 3*i+1] = c[3*i+1];
					vertColors[3*start + 3*i+2] = c[3*i+2];
				}
			}
		//}
		
		//current_size += size;
		
		//if (ray_cast) raycast_size += size;
		
		out_of_sync = true;
	}
	
	/**
	 * PROBLEM: reordering needs to change object properties as well...!
	 */
	public void reorderObjects() {
		
		//alternative with ArrayList,
		//but copy to array needs to maintain an ArrayList and the Array at a point
		
		// 1) keep indices (from-to)
		// 2) take first element - keep or copy
		// 3) if copy: fill end of array
		
		// => iterate from front and back
		// PROBLEM: holes, these would need to be closed!
		// => moving elements as far as necessary
		
		// algorithm:
		// - find first wrong element and swap it with its successor
		// problem: unnecessarily overwrites if already sorted...
		
		System.out.println("ssssssssssssssssize "+object_indices.size() + " array " + vertices.length);
		
		int curr_ind = 0;
		for(int i=0; i<object_indices.size()-1; i++) {
			if (!for_raycast.get(i)) {
				// FIRST COPY ARRAYS
				int first = object_indices.get(i);
				int second = object_indices.get(i+1);
				float[] tmp_v, tmp_t;
				//offset
				int offset = first + second;
				//save the smaller one
				if (first > second) {
					//System.out.println("FIRST IFFF");
					
					tmp_v = new float[3*second];
					tmp_t = new float[2*second];
					
					for (int j=0; j<second;j++) {
						tmp_v[3*j+0] = vertices[3*curr_ind+3*first+3*j+0];
						tmp_v[3*j+1] = vertices[3*curr_ind+3*first+3*j+1];
						tmp_v[3*j+2] = vertices[3*curr_ind+3*first+3*j+2];
						tmp_t[2*j+0] = vertices[2*curr_ind+2*first+2*j+0];
						tmp_t[2*j+1] = vertices[2*curr_ind+2*first+2*j+1];
					}
					//write the larger (first)
					for (int j=0;j<first;j++) {
						//afterwards: very last is first end
						vertices[3*curr_ind + 3*offset - 3*j] = vertices[3*curr_ind + 3*first - 3*j];
						vertices[3*curr_ind + 3*offset - 3*j-1] = vertices[3*curr_ind + 3*first - 3*j-1];
						vertices[3*curr_ind + 3*offset - 3*j-2] = vertices[3*curr_ind + 3*first - 3*j-2];
						vertices[2*curr_ind + 2*offset - 2*j] = vertices[2*curr_ind + 2*first - 2*j];
						vertices[2*curr_ind + 2*offset - 2*j-1] = vertices[2*curr_ind + 2*first - 2*j-1];
					}
					//write the smaller (second)
					for (int j=0;j<second;j++) {
						vertices[3*curr_ind + 3*j] = tmp_v[3*j+0];
						vertices[3*curr_ind + 3*j+1] = tmp_v[3*j+1];
						vertices[3*curr_ind + 3*j+2] = tmp_v[3*j+2];
						vertices[2*curr_ind + 2*j] = tmp_t[2*j+0];
						vertices[2*curr_ind + 2*j+1] = tmp_t[2*j+1];
					}
					
				} else {
					//first smaller
					//System.out.println("SECOND IFFF");
					
					tmp_v = new float[3*first];
					tmp_t = new float[2*first];
					
					for (int j=0; j<first;j++) {
						tmp_v[3*j+0] = vertices[3*curr_ind+3*j+0];
						tmp_v[3*j+1] = vertices[3*curr_ind+3*j+1];
						tmp_v[3*j+2] = vertices[3*curr_ind+3*j+2];
						tmp_t[2*j+0] = vertices[2*curr_ind+2*j+0];
						tmp_t[2*j+1] = vertices[2*curr_ind+2*j+1];
					}
					//write the larger (second)
					for (int j=0;j<second;j++) {
						//afterwards: first end is very last
						vertices[3*curr_ind + 3*j] = vertices[3*curr_ind + 3*(first) + 3*j];
						vertices[3*curr_ind + 3*j+1] = vertices[3*curr_ind + 3*(first) + 3*j+1];
						vertices[3*curr_ind + 3*j+2] = vertices[3*curr_ind + 3*(first) + 3*j+2];
						vertices[2*curr_ind + 2*j] = vertices[2*curr_ind + 2*(first) + 2*j];
						vertices[2*curr_ind + 2*j+1] = vertices[2*curr_ind + 2*(first) + 2*j+1];
					}
					//write the smaller (first)
					for (int j=0;j<first;j++) {
						vertices[3*curr_ind + 3*second + 3*j] = tmp_v[3*j+0];
						vertices[3*curr_ind + 3*second + 3*j+1] = tmp_v[3*j+1];
						vertices[3*curr_ind + 3*second + 3*j+2] = tmp_v[3*j+2];
						vertices[2*curr_ind + 2*second + 2*j] = tmp_t[2*j+0];
						vertices[2*curr_ind + 2*second + 2*j+1] = tmp_t[2*j+1];
					}
					
				}
				// SECOND CHANGE INDEX LISTS
				int tmp = object_indices.get(i+1);
				object_indices.set(i+1, object_indices.get(i));
				object_indices.set(i, tmp);
				
				boolean btmp = for_raycast.get(i+1);
				for_raycast.set(i+1, for_raycast.get(i));
				for_raycast.set(i, btmp);
			}
			curr_ind += object_indices.get(i);
		}
		
	}
	
	public boolean was_modified() {
		return out_of_sync;
	}
	
	public float[] getVertices() {
		out_of_sync = false;
		return vertices;
	}
	
	public float[] getTextCoords() {
		return textCoords;
	}
	
	public float[] getVertColors() {
		return vertColors;
	}
	
	public int getVerticesSize() {
		return current_size;
	}
	
	public int getVerticesSizeForRaycast() {
		return raycast_size;
	}
	
	public void print_vertexdata() {
		for (int i=0; i<current_size; i++) {
			System.out.println("Vertex" +i+ ": "+ vertices[3*i] +"/"+ vertices[3*i+1] +"/"+ vertices[3*i+2]);
		}
	}
	
}